import numpy as np
import xarray as xr
import xradar as xd
from numpy.testing import assert_allclose, assert_almost_equal
from open_radar_data import DATASETS

import pyart

filename = DATASETS.fetch("cfrad.20080604_002217_000_SPOL_v36_SUR.nc")
cfradial2_file = DATASETS.fetch("cfrad2.20080604_002217_000_SPOL_v36_SUR.nc")

COMMON_MAP_TO_GRID_ARGS = {
    "grid_shape": (3, 9, 10),
    "grid_limits": ((-400.0, 400.0), (-900.0, 900.0), (-900, 900)),
    "fields": None,
    "gridding_algo": "map_to_grid",
    "roi_func": lambda z, y, x: 30,
}


def test_get_field(filename=filename):
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = pyart.xradar.Xradar(dtree)
    reflectivity = radar.get_field(0, "DBZ")
    assert reflectivity.shape == (480, 996)


def test_get_azimuth(filename=filename):
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = pyart.xradar.Xradar(dtree)
    azimuths = radar.get_azimuth(0)
    assert azimuths.shape == (480,)


def test_instrument_parameters(filename=filename):
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = pyart.xradar.Xradar(dtree)
    assert radar.instrument_parameters["instrument_name"] == "SPOLRVP8"
    assert_allclose(
        radar.instrument_parameters["latitude"]["data"], np.array(22.52669907)
    )
    assert_allclose(
        radar.instrument_parameters["longitude"]["data"], np.array(120.4335022)
    )


def test_get_gate_x_y_z(filename=filename):
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = pyart.xradar.Xradar(dtree)
    x, y, z = radar.get_gate_x_y_z(0)
    assert x.shape == (480, 996)
    assert y.shape == (480, 996)
    assert z.shape == (480, 996)


def test_get_gate_lat_lon(filename=filename):
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = pyart.xradar.Xradar(dtree)
    lat, lon, alt = radar.get_gate_lat_lon_alt(0)
    # Check lat, lon, and alt values
    assert lat.shape == (480, 996)
    assert_allclose(lat.min(), 21.183521)
    assert lon.shape == (480, 996)
    assert_allclose(lon.min(), 118.97935)
    assert alt.shape == (480, 996)
    assert_allclose(alt.min(), 45.0000017)


def test_add_field(filename=filename):
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = pyart.xradar.Xradar(dtree)
    new_field = radar.fields["DBZ"]
    radar.add_field("reflectivity", new_field)
    assert "reflectivity" in radar.fields
    assert radar["sweep_0"]["reflectivity"].shape == radar["sweep_0"]["DBZ"].shape
    # Test add field like as well
    radar.add_field_like("reflectivity", "test", new_field["data"])
    assert "test" in radar.fields
    assert "data" in radar.fields["test"]
    assert radar.fields["test"]["units"] == "dBZ"


def test_grid(filename=filename):
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = pyart.xradar.Xradar(dtree)
    grid = pyart.map.grid_from_radars(
        (radar,),
        grid_shape=(1, 11, 11),
        grid_limits=((2000, 2000), (-100_000.0, 100_000.0), (-100_000.0, 100_000.0)),
        fields=["DBZ"],
    )
    assert_allclose(grid.x["data"], np.arange(-100_000, 120_000, 20_000))
    assert_allclose(grid.fields["DBZ"]["data"][0, -1, 0], np.array(-0.511), rtol=1e-03)


def _check_attrs_similar(grid1, grid2, attr):
    print("Checking attribute:", attr)
    dic1 = getattr(grid1, attr)
    dic2 = getattr(grid2, attr)
    _check_dicts_similar(dic1, dic2)


def _check_dicts_similar(dic1, dic2):
    for k, v in dic1.items():
        print("Checking key:", k)
        if k == "data":
            assert_almost_equal(v, dic2[k])
        else:
            assert dic2[k] == v


def test_grid_write_read():
    # test the read_grid and write_grid function by performing a
    # write/read roundtrip and comparing the two Grid objects
    grid1 = pyart.testing.make_target_grid()
    grid1.projection["comment"] = "This is a comment"
    grid1.metadata["comment"] = "This is another comment"

    with pyart.testing.InTemporaryDirectory():
        tmpfile = "tmp_grid2.nc"
        pyart.io.write_grid(tmpfile, grid1)
        grid2 = xr.open_dataset(tmpfile, decode_times=False)
        grid2 = pyart.xradar.Xgrid(grid2)

        # check fields
        for field in grid1.fields.keys():
            _check_dicts_similar(grid1.fields[field], grid2.fields[field])

        # check attributes
        assert "Conventions" in grid2.metadata
        grid2.metadata.pop("Conventions")
        _check_attrs_similar(grid1, grid2, "metadata")

        _check_attrs_similar(grid1, grid2, "time")

        _check_attrs_similar(grid1, grid2, "origin_latitude")
        _check_attrs_similar(grid1, grid2, "origin_longitude")
        _check_attrs_similar(grid1, grid2, "origin_altitude")

        _check_attrs_similar(grid1, grid2, "x")
        _check_attrs_similar(grid1, grid2, "y")
        _check_attrs_similar(grid1, grid2, "z")

        _check_attrs_similar(grid1, grid2, "point_x")
        _check_attrs_similar(grid1, grid2, "point_y")
        _check_attrs_similar(grid1, grid2, "point_z")

        _check_attrs_similar(grid1, grid2, "projection")
        assert grid1.projection["_include_lon_0_lat_0"] is True

        _check_attrs_similar(grid1, grid2, "point_latitude")
        _check_attrs_similar(grid1, grid2, "point_longitude")
        _check_attrs_similar(grid1, grid2, "point_altitude")

        assert grid1.nx == grid2.nx
        assert grid1.ny == grid2.ny
        assert grid1.nz == grid2.nz

        _check_attrs_similar(grid1, grid2, "radar_latitude")
        _check_attrs_similar(grid1, grid2, "radar_longitude")
        _check_attrs_similar(grid1, grid2, "radar_altitude")
        _check_attrs_similar(grid1, grid2, "radar_time")
        assert grid1.radar_name["data"] == grid2.radar_name["data"]
        assert grid1.nradar == grid2.nradar
        grid2.ds.close()


def test_to_xradar_object():
    """Test the to_radar_object"""
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = dtree.pyart.to_radar()
    reflectivity = radar.get_field(0, "DBZ")
    assert reflectivity.shape == (480, 996)
    assert radar.nsweeps == 9
    assert radar.ngates == 996


def test_cfradial2_data():
    """
    Test swapping the dimensions when the first is time
    """
    dtree = xr.open_datatree(
        cfradial2_file,
    )

    # Before the transformation, time should be in the dimensions
    assert "time" in dtree["sweep_0"].dims

    # After the transformation, azimuth should be moved from coord to dim
    radar = dtree.pyart.to_radar()
    assert "azimuth" in radar["sweep_0"].dims

    # Ensure georeferencing works as expected
    x, y, z = radar.get_gate_x_y_z(0)
    assert x.shape == (480, 996)


def test_grid_xradar():
    dtree = xd.io.open_cfradial1_datatree(
        filename,
        optional=False,
    )
    radar = dtree.pyart.to_radar()
    grid = pyart.map.grid_from_radars([radar], **COMMON_MAP_TO_GRID_ARGS)
    assert (10, 9, 3) == (grid.nx, grid.ny, grid.nz)
    assert_allclose(np.min(grid.fields["DBZ"]["data"]), -1.3201784)



"""
Corrects polarimetric variables for noise

"""

import warnings

import dask
import numpy as np
import pint
from scipy import signal

from ..config import get_field_name, get_metadata
from ..core.radar import Radar


def correct_noise_rhohv(
    radar,
    urhohv_field=None,
    snr_field=None,
    zdr_field=None,
    nh_field=None,
    nv_field=None,
    rhohv_field=None,
):
    """
    Corrects RhoHV for noise according to eq. 6 in Gourley et al. 2006.
    This correction should only be performed if noise has not been subtracted
    from the signal during the moments computation.

    Parameters
    ----------
    radar : Radar
        Radar object.
    urhohv_field : str, optional
        Name of the RhoHV uncorrected for noise field.
    snr_field, zdr_field, nh_field, nv_field : str, optional
        Names of the SNR, ZDR, horizontal channel noise in dBZ and vertical
        channel noise in dBZ used to correct RhoHV.
    rhohv_field : str, optional
        Name of the rhohv field to output.

    Returns
    -------
    rhohv : dict
        Noise corrected RhoHV field.

    References
    ----------
    Gourley et al. Data Quality of the Meteo-France C-Band Polarimetric
    Radar, JAOT, 23, 1340-1356

    """
    # parse the field parameters
    if urhohv_field is None:
        urhohv_field = get_field_name("uncorrected_cross_correlation_ratio")
    if snr_field is None:
        snr_field = get_field_name("signal_to_noise_ratio")
    if zdr_field is None:
        zdr_field = get_field_name("differential_reflectivity")
    if nh_field is None:
        nh_field = get_field_name("noisedBZ_hh")
    if nv_field is None:
        nv_field = get_field_name("noisedBZ_vv")
    if rhohv_field is None:
        rhohv_field = get_field_name("cross_correlation_ratio")

    # extract fields from radar
    if urhohv_field in radar.fields:
        urhohv = radar.fields[urhohv_field]["data"]
    else:
        raise KeyError("Field not available: " + urhohv_field)
    if snr_field in radar.fields:
        snrdB_h = radar.fields[snr_field]["data"]
    else:
        raise KeyError("Field not available: " + snr_field)
    if zdr_field in radar.fields:
        zdrdB = radar.fields[zdr_field]["data"]
    else:
        raise KeyError("Field not available: " + zdr_field)
    if nh_field in radar.fields:
        nh = radar.fields[nh_field]["data"]
    else:
        raise KeyError("Field not available: " + nh_field)
    if nv_field in radar.fields:
        nv = radar.fields[nv_field]["data"]
    else:
        raise KeyError("Field not available: " + nv_field)

    snr_h = np.ma.power(10.0, 0.1 * snrdB_h)
    zdr = np.ma.power(10.0, 0.1 * zdrdB)
    alpha = np.ma.power(10.0, 0.1 * (nh - nv))

    rhohv_data = urhohv * np.ma.sqrt(
        (1.0 + 1.0 / snr_h) * (1.0 + zdr / (alpha * snr_h))
    )
    rhohv_data[rhohv_data > 1.0] = 1.0

    rhohv = get_metadata(rhohv_field)
    rhohv["data"] = rhohv_data

    return rhohv


def correct_bias(radar, bias=0.0, field_name=None):
    """
    Corrects a radar data bias. If field name is none the correction is
    applied to horizontal reflectivity by default.

    Parameters
    ----------
    radar : Radar
        Radar object.
    bias : float, optional
        The bias magnitude.
    field_name: str, optional
        Names of the field to be corrected.

    Returns
    -------
    corrected_field : dict
        The corrected field

    """
    # parse the field parameters
    if field_name is None:
        field_name = get_field_name("reflectivity")

    # extract fields from radar
    if field_name in radar.fields:
        field_data = radar.fields[field_name]["data"]
    else:
        raise KeyError("Field not available: " + field_name)

    corr_field_data = field_data - bias

    if field_name.startswith("corrected_"):
        corr_field_name = field_name
    else:
        corr_field_name = "corrected_" + field_name

    corr_field = get_metadata(corr_field_name)
    corr_field["data"] = corr_field_data

    return corr_field


def calc_zdr_offset(radar, gatefilter=None, height_range=None, zdr_var=None):
    """
    Function for calculating the ZDR bias from a VPT scan.

    Parameters
    ----------
    radar : PyART radar object
        Radar object with radar data
    gatefilter: PyART GateFilter
        Gatefilter for filtering out data for calculating ZDR bias
    height_range: tuple
        The minimum and maximum heights in meters for the scan.
    zdr_var: string or None
        The name of the ZDR variable. Set to None to have PyART try to determine this automatically.

    Returns
    -------
    profiles : dict
        The mean vertical profiles of each radar moment are extracted along with the ZDR bias.

    """
    if height_range is None:
        height_range = (0, 100000.0)

    if zdr_var is None:
        zdr_var = get_field_name("differential_reflectivity")

    height_mask = np.logical_and(
        radar.range["data"] >= height_range[0], radar.range["data"] <= height_range[1]
    )

    mask = gatefilter.gate_included
    Zdr = radar.fields[zdr_var]["data"]
    if isinstance(Zdr, np.ma.MaskedArray):
        Zdr = Zdr.filled(np.nan)
    Zdr = np.where(mask, Zdr, np.nan)
    bias = np.nanmean(Zdr[:, height_mask])
    height_mask_tiled = np.tile(height_mask, (radar.nrays, 1))
    Zdr = np.where(height_mask_tiled, Zdr, np.nan)
    results = {
        "bias": bias,
        "profile_zdr": np.nanmean(Zdr[:, :], axis=0),
        "range": radar.range["data"],
    }
    for k in radar.fields.keys():
        if k != "range":
            field_data = radar.fields[k]["data"].astype(float)
            if isinstance(field_data, np.ma.MaskedArray):
                field_data = field_data.filled(np.nan)

            field_data = np.where(mask, field_data, np.nan)
            field_data = np.where(height_mask_tiled, field_data, np.nan)
            results["profile_" + k] = np.nanmean(field_data[:, :], axis=0)
    return results


def calc_cloud_mask(
    radar,
    field,
    height=None,
    noise_threshold=-45.0,
    threshold_offset=5.0,
    counts_threshold=12,
):
    """
    Primary function for calculating the cloud mask.

    Parameters
    ----------
    radar : Radar
        Py-ART Radar object.
    field : string
        Reflectivity field name to calculate.
    height : string
        Height name to use for calculations.
    noise_threshold : float
        Threshold value used for noise detection. Greater than this value.
    threshold_offset : float
        Threshold offset value used for noise detection
    counts_threshold : int
        Threshold of counts used to determine mask. Greater than or equal to this value.

    Returns
    -------
    radar : Radar
        Returns an updated Radar object with cloud mask fields.

    References
    ----------
    Kollias, P., I. Jo, P. Borque, A. Tatarevic, K. Lamer, N. Bharadwaj, K. Widener,
    K. Johnson, and E.E. Clothiaux, 2014: Scanning ARM Cloud Radars. Part II: Data
    Quality Control and Processing. J. Atmos. Oceanic Technol., 31, 583–598,
    https://doi.org/10.1175/JTECH-D-13-00045.1

    """

    if not isinstance(radar, Radar):
        raise ValueError("Please use a valid Py-ART Radar object")

    if not isinstance(field, str):
        raise ValueError("Please specify a valid field name.")

    noise = calc_noise_floor(radar, field, height=height)

    noise_thresh = (
        np.nanmin(
            np.vstack(
                [
                    noise,
                    np.full(np.shape(radar.fields[field]["data"])[0], noise_threshold),
                ]
            ),
            axis=0,
        )
        + threshold_offset
    )

    data = range_correction(radar, field, height=height)

    task = []
    for i in range(np.shape(data)[0]):
        task.append(dask.delayed(_first_mask)(data[i, :], noise_thresh[i]))

    result = dask.compute(task)
    mask1 = np.array(result[0])

    counts = signal.convolve2d(mask1, np.ones((4, 4), dtype=int), mode="same")
    mask2 = np.zeros_like(data, dtype=np.int16)
    mask2[counts >= counts_threshold] = 1

    cloud_mask_1 = {}
    cloud_mask_1["long_name"] = "Cloud mask 1 (linear profile)"
    cloud_mask_1["units"] = "1"
    cloud_mask_1["comment"] = (
        "The mask is calculated with a " "linear mask along each time profile."
    )
    cloud_mask_1["flag_values"] = [0, 1]
    cloud_mask_1["flag_meanings"] = ["no_cloud", "cloud"]
    cloud_mask_1["data"] = mask1

    cloud_mask_2 = {}
    cloud_mask_2["long_name"] = "Cloud mask 2 (2D box)"
    cloud_mask_2["units"] = "1"
    cloud_mask_2["comment"] = "The mask uses a 2D box to " "filter out noise."
    cloud_mask_2["flag_values"] = [0, 1]
    cloud_mask_2["flag_meanings"] = ["no_cloud", "cloud"]
    cloud_mask_2["data"] = mask2

    radar.add_field("cloud_mask_1", cloud_mask_1, replace_existing=True)
    radar.add_field("cloud_mask_2", cloud_mask_2, replace_existing=True)

    return radar


def calc_noise_floor(radar, field, height):
    """
    Calculation for getting the noise floor

    Parameters
    ----------
    radar : Radar
        Py-ART Radar object containing data.
    field : string
        Reflectivity field name to correct.
    height : string
        Height name to use in correction.

    Returns
    -------
    noise : array
        Returns the noise floor value for each time sample.

    References
    ----------
    Kollias, P., I. Jo, P. Borque, A. Tatarevic, K. Lamer, N. Bharadwaj, K. Widener,
    K. Johnson, and E.E. Clothiaux, 2014: Scanning ARM Cloud Radars. Part II: Data
    Quality Control and Processing. J. Atmos. Oceanic Technol., 31, 583–598,
    https://doi.org/10.1175/JTECH-D-13-00045.1

    """

    if not isinstance(radar, Radar):
        raise ValueError("Please use a valid Py-ART Radar object.")

    # Range correct data and return the array from the Radar object
    data = range_correction(radar, field, height=height)

    # Pass each timestep into task list to calculate cloud threshhold
    # with a delayed dask process
    task = [dask.delayed(cloud_threshold)(row) for row in data]

    # Perform dask computation
    noise = dask.compute(*task)

    # Convert returned dask tuple into numpy array
    noise = np.array(noise, dtype=float)

    return noise


def cloud_threshold(data, n_avg=1.0, nffts=None):
    """
    Calculates the noise floor from a cloud threshold.

    Parameters
    ----------
    data : array
        Numpy array
    n_avg : float
        Number of points to average over
    nffts : int
        Number of heights to iterate over. If None will use the size of data.

    Returns
    -------
    result : numpy scalar float
        Returns the noise floor value for each time sample.

    References
    ----------
    Kollias, P., I. Jo, P. Borque, A. Tatarevic, K. Lamer, N. Bharadwaj, K. Widener,
    K. Johnson, and E.E. Clothiaux, 2014: Scanning ARM Cloud Radars. Part II: Data
    Quality Control and Processing. J. Atmos. Oceanic Technol., 31, 583–598,
    https://doi.org/10.1175/JTECH-D-13-00045.1

    """

    if nffts is None:
        nffts = data.size

    data = 10.0 ** (data / 10.0)
    data = np.sort(data)

    nthld = 10.0**-10.0
    dsum = 0.0
    sumSq = 0.0
    n = 0.0
    numNs = []
    sqrt_n_avg = np.sqrt(n_avg)
    for i in range(nffts):
        if data[i] > nthld:
            dsum += data[i]
            sumSq += data[i] ** 2.0
            n += 1.0
            a3 = dsum * dsum
            a1 = sqrt_n_avg * (n * sumSq - a3)
            if n > nffts / 4.0:
                if a1 <= a3:
                    sumNs = dsum
                    numNs = [n]
            else:
                sumNs = dsum
                numNs = [n]

    if len(numNs) > 0:
        n_mean = sumNs / numNs[0]
    else:
        n_mean = np.nan

    if n_mean == 0.0:
        result = np.nan
    else:
        result = 10.0 * np.log10(n_mean)

    return result


def range_correction(radar, field, height):
    """
    Corrects reflectivity for range to help get the
    correct noise floor values

    Parameters
    ----------
    radar : Radar
        Py-ART Radar object containing data.
    field : string
        Reflectivity field name to correct.
    height : string
        Height name to use in correction.

    Returns
    -------
    data : array
        Returns a range corrected array matching reflectivity field.

    """

    if not isinstance(radar, Radar):
        raise ValueError("Please use a valid Py-ART Radar object.")

    try:
        height_units = getattr(radar, height)["units"]
    except KeyError:
        warnings.warn(
            f"Height '{height} does not have units attribute. "
            "Assuming units are meters."
        )
        height_units = "m"

    height = getattr(radar, height)["data"]
    desired_unit = "m"
    if height_units is not desired_unit:
        if isinstance(height, np.ma.MaskedArray):
            height = height.filled(np.nan)
        unit_registry = pint.UnitRegistry()
        height = height * unit_registry.parse_expression(height_units)
        height = height.to(desired_unit)
        height = height.magnitude

    data = radar.fields[field]["data"]

    if isinstance(data, np.ma.MaskedArray) and not data.mask:
        data = data.data
    elif isinstance(data, np.ma.MaskedArray) and data.mask:
        data = data.filled(np.nan)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", category=RuntimeWarning, message=".*divide by zero encountered.*"
        )
        data = data - 20.0 * np.log10(height / 1000.0)

    return data


def _first_mask(data, noise_threshold):
    mask = np.zeros_like(data, dtype=np.int16)
    mask[data > noise_threshold] = 1
    return mask



"""
Constant Altitude Plan Position Indicator
"""

import numpy as np
from netCDF4 import num2date
from pandas import to_datetime
from scipy.interpolate import RectBivariateSpline

from pyart.core import Radar


def create_cappi(
    radar,
    fields=None,
    height=2000,
    gatefilter=None,
    vel_field="velocity",
    same_nyquist=True,
    nyquist_vector_idx=0,
):
    """
    Create a Constant Altitude Plan Position Indicator (CAPPI) from radar data.

    Parameters
    ----------
    radar : Radar
        Py-ART Radar object containing the radar data.
    fields : list of str, optional
        List of radar fields to be used for creating the CAPPI.
        If None, all available fields will be used. Default is None.
    height : float, optional
        The altitude at which to create the CAPPI. Default is 2000 meters.
    gatefilter : GateFilter, optional
        A GateFilter object to apply masking/filtering to the radar data.
        Default is None.
    vel_field : str, optional
        The name of the velocity field to be used for determining the Nyquist velocity.
        Default is 'velocity'.
    same_nyquist : bool, optional
        Whether to only stack sweeps with the same Nyquist velocity.
        Default is True.
    nyquist_vector_idx : int, optional
        Index for the Nyquist velocity vector if `same_nyquist` is True.
        Default is 0.

    Returns
    -------
    Radar
        A Py-ART Radar object containing the CAPPI at the specified height.

    Notes
    -----
    CAPPI (Constant Altitude Plan Position Indicator) is a radar visualization
    technique that provides a horizontal view of meteorological data at a fixed altitude.
    Reference: https://glossary.ametsoc.org/wiki/Cappi

    Author
    ------
    Hamid Ali Syed (@syedhamidali)
    """

    if fields is None:
        fields = list(radar.fields.keys())

    # Initialize the first sweep as the reference
    first_sweep = 0

    # Initialize containers for the stacked data and nyquist velocities
    data_stack = []
    nyquist_stack = []

    # Process each sweep individually
    for sweep in range(radar.nsweeps):
        sweep_slice = radar.get_slice(sweep)
        try:
            nyquist = radar.get_nyquist_vel(sweep=sweep)
            nyquist = np.round(nyquist)
        except LookupError:
            print(
                f"Nyquist velocity unavailable for sweep {sweep}. Estimating using maximum velocity."
            )
            nyquist = radar.fields[vel_field]["data"][sweep_slice].max()

        sweep_data = {}

        for field in fields:
            data = radar.get_field(sweep, field)

            # Apply gatefilter if provided
            if gatefilter is not None:
                data = np.ma.masked_array(
                    data, gatefilter.gate_excluded[sweep_slice, :]
                )
            time = radar.time["data"][sweep_slice]

            # Extract and sort azimuth angles
            azimuth = radar.azimuth["data"][sweep_slice]
            azimuth_sorted_idx = np.argsort(azimuth)
            azimuth = azimuth[azimuth_sorted_idx]
            data = data[azimuth_sorted_idx]

            # Store initial lat/lon for reordering
            if sweep == first_sweep:
                azimuth_final = azimuth
                time_final = time
            else:
                # Interpolate data for consistent azimuth ordering across sweeps
                interpolator = RectBivariateSpline(azimuth, radar.range["data"], data)
                data = interpolator(azimuth_final, radar.range["data"])

            sweep_data[field] = data[np.newaxis, :, :]

        data_stack.append(sweep_data)
        nyquist_stack.append(nyquist)

    nyquist_stack = np.array(nyquist_stack)

    # Filter for sweeps with similar Nyquist velocities
    if same_nyquist:
        nyquist_range = nyquist_stack[nyquist_vector_idx]
        nyquist_mask = np.abs(nyquist_stack - nyquist_range) <= 1
        data_stack = [
            sweep_data for i, sweep_data in enumerate(data_stack) if nyquist_mask[i]
        ]

    # Generate CAPPI for each field using data_stack
    fields_data = {}
    for field in fields:
        data_3d = np.concatenate(
            [sweep_data[field] for sweep_data in data_stack], axis=0
        )

        # Sort azimuth for all sweeps
        dim0 = data_3d.shape[1:]
        azimuths = np.linspace(0, 359, dim0[0])
        elevation_angles = radar.fixed_angle["data"][: data_3d.shape[0]]
        ranges = radar.range["data"]

        theta = (450 - azimuths) % 360
        THETA, PHI, R = np.meshgrid(theta, elevation_angles, ranges)
        Z = R * np.sin(PHI * np.pi / 180)

        # Extract the data slice corresponding to the requested height
        height_idx = np.argmin(np.abs(Z - height), axis=0)
        CAPPI = np.array(
            [
                data_3d[height_idx[j, i], j, i]
                for j in range(dim0[0])
                for i in range(dim0[1])
            ]
        ).reshape(dim0)

        # Retrieve units and handle case where units might be missing
        units = radar.fields[field].get("units", "").lower()

        # Determine valid_min and valid_max based on units
        if units == "dbz":
            valid_min, valid_max = -10, 80
        elif units in ["m/s", "meters per second"]:
            valid_min, valid_max = -100, 100
        elif units == "db":
            valid_min, valid_max = -7.9, 7.9
        else:
            # If units are not found or don't match known types, set default values or skip masking
            valid_min, valid_max = None, None

        # If valid_min or valid_max are still None, set them to conservative defaults or skip
        if valid_min is None:
            print(f"Warning: valid_min not set for {field}, using default of -1e10")
            valid_min = -1e10  # Conservative default
        if valid_max is None:
            print(f"Warning: valid_max not set for {field}, using default of 1e10")
            valid_max = 1e10  # Conservative default

        # Apply valid_min and valid_max masking
        if valid_min is not None:
            CAPPI = np.ma.masked_less(CAPPI, valid_min)
        if valid_max is not None:
            CAPPI = np.ma.masked_greater(CAPPI, valid_max)

        # Convert to masked array with the specified fill value
        CAPPI.set_fill_value(radar.fields[field].get("_FillValue", np.nan))
        CAPPI = np.ma.masked_invalid(CAPPI)
        CAPPI = np.ma.masked_outside(CAPPI, valid_min, valid_max)

        fields_data[field] = {
            "data": CAPPI,
            "units": radar.fields[field]["units"],
            "long_name": f"CAPPI {field} at {height} meters",
            "comment": f"CAPPI {field} calculated at a height of {height} meters",
            "_FillValue": radar.fields[field].get("_FillValue", np.nan),
        }

    # Set the elevation to zeros for CAPPI
    elevation_final = np.zeros(dim0[0], dtype="float32")

    # Since we are using the whole volume scan, report mean time
    try:
        dtime = to_datetime(
            num2date(radar.time["data"], radar.time["units"]).astype(str),
            format="ISO8601",
        )
    except ValueError:
        dtime = to_datetime(
            num2date(radar.time["data"], radar.time["units"]).astype(str)
        )
    dtime = dtime.mean()

    time = radar.time.copy()
    time["data"] = time_final
    time["mean"] = dtime

    # Create the Radar object with the new CAPPI data
    return Radar(
        time=radar.time.copy(),
        _range=radar.range.copy(),
        fields=fields_data,
        metadata=radar.metadata.copy(),
        scan_type=radar.scan_type,
        latitude=radar.latitude.copy(),
        longitude=radar.longitude.copy(),
        altitude=radar.altitude.copy(),
        sweep_number=radar.sweep_number.copy(),
        sweep_mode=radar.sweep_mode.copy(),
        fixed_angle=radar.fixed_angle.copy(),
        sweep_start_ray_index=radar.sweep_start_ray_index.copy(),
        sweep_end_ray_index=radar.sweep_end_ray_index.copy(),
        azimuth=radar.azimuth.copy(),
        elevation={"data": elevation_final},
        instrument_parameters=radar.instrument_parameters,
    )



"""
==========================================================
Read and Plot Cfradial2/FM301 data Using Xradar and Py-ART
==========================================================

An example which uses xradar and Py-ART to read and plot Cfradial2/FM301 data.

"""

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause


import xarray as xr
from open_radar_data import DATASETS

import pyart

# Locate the test data and read in using xradar
filename = DATASETS.fetch("cfrad2.20080604_002217_000_SPOL_v36_SUR.nc")
tree = xr.open_datatree(filename)

# Give the tree Py-ART radar methods
radar = tree.pyart.to_radar()

# Plot the Reflectivity Field (corrected_reflectivity_horizontal)
display = pyart.graph.RadarMapDisplay(radar)
display.plot_ppi("DBZ", cmap="ChaseSpectral", vmin=-20, vmax=70)



""" Unit Tests for Py-ART's retrieve/echo_class.py module. """

import numpy as np

import pyart


def test_cfad_default():
    # initalize test radar object
    radar = pyart.io.read(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)
    ref_field = "reflectivity"

    # set every value to 20
    radar.fields[ref_field]["data"] = (
        np.ones(radar.fields[ref_field]["data"].shape) * 20
    )

    # set mask to none
    field_mask = np.zeros(radar.fields[ref_field]["data"].shape)

    # calculate CFAD
    freq_norm, height_edges, field_edges = pyart.retrieve.create_cfad(
        radar,
        field_bins=np.linspace(0, 30, 20),
        altitude_bins=np.arange(0, 18000, 100),
        field="reflectivity",
        field_mask=field_mask,
    )

    # set row to mask and test
    verify_index = 12

    # if CFAD code works correctly, each column should have the same values and only 1 column should have a value of 1
    # check all columns are the same
    assert freq_norm.all(axis=0).any()
    # check column 12 is all ones
    assert (freq_norm[:, verify_index] == 1).all()



import numpy as np
import scipy.ndimage


def _steiner_conv_strat(
    refl,
    x,
    y,
    dx,
    dy,
    intense=42,
    peak_relation=0,
    area_relation=1,
    bkg_rad=11000,
    use_intense=True,
):
    """
    We perform the Steiner et al. (1995) algorithm for echo classification
    using only the reflectivity field in order to classify each grid point
    as either convective, stratiform or undefined. Grid points are
    classified as follows,

    0 = Undefined
    1 = Stratiform
    2 = Convective
    """

    def convective_radius(ze_bkg, area_relation):
        """
        Given a mean background reflectivity value, we determine via a step
        function what the corresponding convective radius would be.

        Higher background reflectivitives are expected to have larger
        convective influence on surrounding areas, so a larger convective
        radius would be prescribed.
        """
        if area_relation == 0:
            if ze_bkg < 30:
                conv_rad = 1000.0
            elif (ze_bkg >= 30) & (ze_bkg < 35.0):
                conv_rad = 2000.0
            elif (ze_bkg >= 35.0) & (ze_bkg < 40.0):
                conv_rad = 3000.0
            elif (ze_bkg >= 40.0) & (ze_bkg < 45.0):
                conv_rad = 4000.0
            else:
                conv_rad = 5000.0

        if area_relation == 1:
            if ze_bkg < 25:
                conv_rad = 1000.0
            elif (ze_bkg >= 25) & (ze_bkg < 30.0):
                conv_rad = 2000.0
            elif (ze_bkg >= 30.0) & (ze_bkg < 35.0):
                conv_rad = 3000.0
            elif (ze_bkg >= 35.0) & (ze_bkg < 40.0):
                conv_rad = 4000.0
            else:
                conv_rad = 5000.0

        if area_relation == 2:
            if ze_bkg < 20:
                conv_rad = 1000.0
            elif (ze_bkg >= 20) & (ze_bkg < 25.0):
                conv_rad = 2000.0
            elif (ze_bkg >= 25.0) & (ze_bkg < 30.0):
                conv_rad = 3000.0
            elif (ze_bkg >= 30.0) & (ze_bkg < 35.0):
                conv_rad = 4000.0
            else:
                conv_rad = 5000.0

        if area_relation == 3:
            if ze_bkg < 40:
                conv_rad = 0.0
            elif (ze_bkg >= 40) & (ze_bkg < 45.0):
                conv_rad = 1000.0
            elif (ze_bkg >= 45.0) & (ze_bkg < 50.0):
                conv_rad = 2000.0
            elif (ze_bkg >= 50.0) & (ze_bkg < 55.0):
                conv_rad = 6000.0
            else:
                conv_rad = 8000.0

        return conv_rad

    def peakedness(ze_bkg, peak_relation):
        """
        Given a background reflectivity value, we determine what the necessary
        peakedness (or difference) has to be between a grid point's
        reflectivity and the background reflectivity in order for that grid
        point to be labeled convective.
        """
        if peak_relation == 0:
            if ze_bkg < 0.0:
                peak = 10.0
            elif (ze_bkg >= 0.0) and (ze_bkg < 42.43):
                peak = 10.0 - ze_bkg**2 / 180.0
            else:
                peak = 0.0

        elif peak_relation == 1:
            if ze_bkg < 0.0:
                peak = 14.0
            elif (ze_bkg >= 0.0) and (ze_bkg < 42.43):
                peak = 14.0 - ze_bkg**2 / 180.0
            else:
                peak = 4.0

        return peak

    sclass = np.zeros(refl.shape, dtype=int)
    ny, nx = refl.shape

    for i in range(0, nx):
        # Get stencil of x grid points within the background radius
        imin = np.max(np.array([1, (i - bkg_rad / dx)], dtype=int))
        imax = np.min(np.array([nx, (i + bkg_rad / dx)], dtype=int))

        for j in range(0, ny):
            # First make sure that the current grid point has not already been
            # classified. This can happen when grid points within the
            # convective radius of a previous grid point have also been
            # classified.
            if ~np.isnan(refl[j, i]) & (sclass[j, i] == 0):
                # Get stencil of y grid points within the background radius
                jmin = np.max(np.array([1, (j - bkg_rad / dy)], dtype=int))
                jmax = np.min(np.array([ny, (j + bkg_rad / dy)], dtype=int))

                n = 0
                sum_ze = 0

                # Calculate the mean background reflectivity for the current
                # grid point, which will be used to determine the convective
                # radius and the required peakedness.

                for r in range(imin, imax):
                    for m in range(jmin, jmax):
                        if not np.isnan(refl[m, r]):
                            rad = np.sqrt((x[r] - x[i]) ** 2 + (y[m] - y[j]) ** 2)

                            # The mean background reflectivity will first be
                            # computed in linear units, i.e. mm^6/m^3, then
                            # converted to decibel units.
                            if rad <= bkg_rad:
                                n += 1
                                sum_ze += 10.0 ** (refl[m, r] / 10.0)

                if n == 0:
                    ze_bkg = np.inf
                else:
                    ze_bkg = 10.0 * np.log10(sum_ze / n)

                # Now get the corresponding convective radius knowing the mean
                # background reflectivity.
                conv_rad = convective_radius(ze_bkg, area_relation)

                # Now we want to investigate the points surrounding the current
                # grid point that are within the convective radius, and whether
                # they too are convective, stratiform or undefined.

                # Get stencil of x and y grid points within the convective
                # radius.
                lmin = np.max(np.array([1, int(i - conv_rad / dx)], dtype=int))
                lmax = np.min(np.array([nx, int(i + conv_rad / dx)], dtype=int))
                mmin = np.max(np.array([1, int(j - conv_rad / dy)], dtype=int))
                mmax = np.min(np.array([ny, int(j + conv_rad / dy)], dtype=int))

                if use_intense and (refl[j, i] >= intense):
                    sclass[j, i] = 2

                    for r in range(lmin, lmax):
                        for m in range(mmin, mmax):
                            if not np.isnan(refl[m, r]):
                                rad = np.sqrt((x[r] - x[i]) ** 2 + (y[m] - y[j]) ** 2)

                                if rad <= conv_rad:
                                    sclass[m, r] = 2

                else:
                    peak = peakedness(ze_bkg, peak_relation)

                    if refl[j, i] - ze_bkg >= peak:
                        sclass[j, i] = 2

                        for r in range(imin, imax):
                            for m in range(jmin, jmax):
                                if not np.isnan(refl[m, r]):
                                    rad = np.sqrt(
                                        (x[r] - x[i]) ** 2 + (y[m] - y[j]) ** 2
                                    )

                                    if rad <= conv_rad:
                                        sclass[m, r] = 2

                    else:
                        # If by now the current grid point has not been
                        # classified as convective by either the intensity
                        # criteria or the peakedness criteria, then it must be
                        # stratiform.
                        sclass[j, i] = 1

    return sclass


def steiner_class_buff(
    ze,
    x,
    y,
    z,
    dx,
    dy,
    bkg_rad,
    work_level,
    intense,
    peak_relation,
    area_relation,
    use_intense,
):
    zslice = np.argmin(np.abs(z - work_level))
    refl = ze[zslice, :, :]

    area_rel = {"small": 0, "medium": 1, "large": 2, "sgp": 3}
    peak_rel = {"default": 0, "sgp": 1}

    sclass = _steiner_conv_strat(
        refl,
        x,
        y,
        dx,
        dy,
        intense=intense,
        peak_relation=peak_rel[peak_relation],
        area_relation=area_rel[area_relation],
        bkg_rad=11000,
        use_intense=True,
    )

    return sclass


def _feature_detection(
    field,
    dx,
    dy,
    always_core_thres=42,
    bkg_rad_km=11,
    use_cosine=True,
    max_diff=5,
    zero_diff_cos_val=55,
    scalar_diff=1.5,
    use_addition=True,
    calc_thres=0.75,
    weak_echo_thres=5.0,
    min_val_used=5.0,
    dB_averaging=True,
    remove_small_objects=True,
    min_km2_size=10,
    binary_close=False,
    val_for_max_rad=30,
    max_rad_km=5.0,
    core_val=3,
    nosfcecho=0,
    weakecho=3,
    bkgd_val=1,
    feat_val=2,
):
    """
    These functions are used to detect features in fields based on how distinct they are from the background
    average. Methodology described in Tomkins et al. (2023), originally based on convective-stratiform algorithms
    developed by Steiner et al. (1995), Yuter and Houze (1997), and Yuter et al. (2005) Grid points are
    classified as follows,

    nosfcecho = No Surface Echo/ Undefined
    bkgd_val = Background echo (e.g. Stratiform)
    feat_val = Feature (e.g. Convective)
    weakecho = Weak Echo

    field : array
        array of values to find features
    x, y : array
        x and y coordinates of field array, respectively
    dx, dy : float
        The x- and y-dimension resolutions in meters, respectively.
    always_core_thres : float, optional
        Threshold for points that are always features. All values above the threshold are classified as features.
    bkg_rad_km : float, optional
        Radius to compute background field in kilometers. Default is 11 km. Recommended to be at least 3 x grid spacing
    use_cosine : bool, optional
        Boolean used to determine if a cosine scheme (see Yuter and Houze (1997)) should be used for identifying
        cores (True) or if a simpler scalar scheme (False) should be used.
    max_diff : float, optional
        Maximum difference between background average and grid value in order to be classified as features.
        "a" value in Eqn. B1 in Yuter and Houze (1997)
    zero_diff_cos_val : float, optional
        Value where difference between background average and grid value is zero in the cosine function
        "b" value in Eqn. B1 in Yuter and Houze (1997)
    scalar_diff : float, optional
        If using a scalar difference scheme, this value is the multiplier or addition to the background average
    use_addition : bool, optional
        Determines if a multiplier (False) or addition (True) in the scalar difference scheme should be used
    calc_thres : float, optional
        Minimum percentage of points needed to be considered in background average calculation
    weak_echo_thres : float, optional
        Threshold for determining weak echo. All values below this threshold will be considered weak echo
    min_val_used : float, optional
        Minimum value used for classification. All values below this threshold will be considered no surface echo
        See Yuter and Houze (1997) and Yuter et al. (2005) for more detail. Units based on input field
    dB_averaging : bool, optional
        True if using dBZ values that need to be converted to linear Z before averaging. False for other types of values
    remove_small_objects : bool, optional
        Determines if small objects should be removed from core array. Default is True.
    min_km2_size : float, optional
        Minimum size of Cores to be considered. Cores less than this size will be removed. Default is 10 km^2.
    binary_close : bool, optional
        Determines if a binary closing should be performed on the cores. Default is False.
    val_for_max_rad : float, optional
        value used for maximum radius. Cores with values above this will have the maximum radius incorporated.
    max_rad_km : float, optional
        Maximum radius around cores to classify as feature. Default is 5 km
    core_val : int, optional
        Value for points classified as cores
    nosfcecho : int, optional
        Value for points classified as no surface echo, based on min_val_used
    weakecho : int, optional
        Value for points classified as weak echo, based on weak_echo_thres.
    bkgd_val : int, optional
        Value for points classified as background echo.
    feat_val : int, optional
        Value for points classified as features.

    Returns
    -------
    field_bkg : array
        Array of background values
    core_array : array
        Array of initial cores (identified features without radii applied)
    feature_array : array
        Array of feature detection with radii applied
    """

    # Set up mask arrays for background average and prepare for mask arrays
    # calculate maximum diameter from max. radius (input)
    max_diameter = int(np.floor((max_rad_km / (dx / 1000)) * 2))
    # if diameter is even, make odd
    if max_diameter % 2 == 0:
        max_diameter = max_diameter + 1
    # find center point
    center_mask_x = int(np.floor(max_diameter / 2))

    # prepare background mask array for computing background average
    # calculate number of pixels for background array given requested background radius and dx
    bkg_diameter_pix = int(np.floor((bkg_rad_km / (dx / 1000)) * 2))
    # set diameter to odd if even
    if bkg_diameter_pix % 2 == 0:
        bkg_diameter_pix = bkg_diameter_pix + 1
    # find center point
    bkg_center = int(np.floor(bkg_diameter_pix / 2))
    # create background array
    bkg_mask_array = np.ones((bkg_diameter_pix, bkg_diameter_pix), dtype=float)
    # mask outside circular region
    bkg_mask_array = create_radial_mask(
        bkg_mask_array,
        min_rad_km=0,
        max_rad_km=bkg_rad_km,
        x_pixsize=dx / 1000,
        y_pixsize=dy / 1000,
        center_x=bkg_center,
        center_y=bkg_center,
        circular=True,
    )

    # Feature detection
    # start by making field a masked array
    field = np.ma.masked_invalid(field)
    # Compute background radius
    field_bkg = calc_bkg_intensity(field, bkg_mask_array, dB_averaging, calc_thres)

    # Get core array from cosine scheme, or scalar scheme
    if use_cosine:
        core_array = core_cos_scheme(
            field, field_bkg, max_diff, zero_diff_cos_val, always_core_thres, core_val
        )
    else:
        core_array = core_scalar_scheme(
            field,
            field_bkg,
            scalar_diff,
            always_core_thres,
            core_val,
            use_addition=use_addition,
        )

    # Assign radii based on background field
    radius_array_km = assign_feature_radius_km(
        field_bkg, val_for_max_rad=val_for_max_rad, max_rad=max_rad_km
    )

    # remove small objects in core array
    if remove_small_objects:
        # calculate minimum pixel size given dx and dy
        min_pix_size = min_km2_size / ((dx / 1000) * (dy / 1000))
        # label connected objects in core array
        cc_labels, _ = scipy.ndimage.label(core_array)
        # mask labels where core array is masked
        cc_labels = np.ma.masked_where(core_array.mask, cc_labels)

        # loop through each unique label
        for lab in np.unique(cc_labels):
            # calculate number of pixels for each label
            size_lab = np.count_nonzero(cc_labels == lab)
            # if number of pixels is less than minimum, then remove core
            if size_lab < min_pix_size:
                core_array[cc_labels == lab] = 0

    # perform binary closing
    if binary_close:
        # binary closing - returns binary array
        close_core = scipy.ndimage.binary_closing(core_array).astype(int)
        # set values to core values
        core_array = close_core * core_val

    # Incorporate radius using binary dilation
    # Create empty array for assignment
    temp_assignment = np.zeros_like(core_array)

    # Loop through radii
    for radius in np.arange(1, max_rad_km + 1):
        # create mask array for radius incorporation
        radius_mask_array = create_radius_mask(
            max_diameter, radius, dx / 1000, dy / 1000, center_mask_x
        )
        # find location of radius
        temp = radius_array_km == radius
        # get cores for given radius
        temp_core = np.ma.masked_where(~temp, core_array)
        # dilate cores
        temp_dilated = scipy.ndimage.binary_dilation(
            temp_core.filled(0), radius_mask_array
        )
        # add to assignment array
        temp_assignment = temp_assignment + temp_dilated

    # add dilated cores to original array
    core_copy = np.ma.copy(core_array)
    core_copy[temp_assignment >= 1] = core_val

    # Now do feature detection
    feature_array = np.zeros_like(field)
    feature_array = classify_feature_array(
        field,
        feature_array,
        core_copy,
        nosfcecho,
        feat_val,
        bkgd_val,
        weakecho,
        core_val,
        min_val_used,
        weak_echo_thres,
    )
    # mask where field is masked
    feature_array = np.ma.masked_where(field.mask, feature_array)

    return field_bkg, core_array, feature_array


# functions


def create_radial_mask(
    mask_array,
    min_rad_km,
    max_rad_km,
    x_pixsize,
    y_pixsize,
    center_x,
    center_y,
    circular=True,
):
    """
    Computes a radial distance mask, everything with distance between minradiuskm
    and maxradiuskm is assigned 1, everything else is assigned 0. This version can
    handle rectangular arrays and pixels as well as square ones.

    Parameters
    ----------
    mask_array : array
        Array to mask
    min_rad_km, max_rad_km : float
        The minimum and maximum radius of the non-masked region in kilometers.
    x_pixsize, y_pixsize : float
        The pixel size in the x- and y-dimension in kilometers, respectively
    center_x, center_y : int
        The center pixel in the x- and y-dimension, respectively
    circular : bool
        True returns circular mask, False returns a rectangular mask.

    Returns
    -------
    mask_array : array
        Rectangular array masked by a radial distance.
    """

    xsize, ysize = mask_array.shape

    for j in np.arange(0, ysize, 1):
        for i in np.arange(0, xsize, 1):
            # compute range to pixel
            if circular:
                x_range_sq = ((center_x - i) * x_pixsize) ** 2
                y_range_sq = ((center_y - j) * y_pixsize) ** 2
                circ_range = np.sqrt(x_range_sq + y_range_sq)
            # if circular is False, use square mask
            else:
                x_range = abs(int(np.floor(center_x - i) * x_pixsize))
                y_range = abs(int(np.floor(center_y - j) * y_pixsize))

                if x_range > y_range:
                    circ_range = x_range
                else:
                    circ_range = y_range
            # if range is within min and max, set to True
            if (circ_range <= max_rad_km) and (circ_range >= min_rad_km):
                mask_array[j, i] = 1
            else:
                mask_array[j, i] = 0

    return mask_array


def calc_bkg_intensity(field, bkg_mask_array, dB_averaging, calc_thres=None):
    """
    Calculate the background of the given field. The footprint used to
    calculate the average for each pixel is given by bkg_mask_array

    Parameters
    ----------
    field : array
        Field array to compute average
    bkg_mask_array : array
        Array of radial points to use for average
    dB_averaging : bool
        If True, converts dBZ to linear Z before averaging
    calc_thres : float
        Minimum percentage of points needed to be considered in background average calculation

    Returns
    -------
    field_bkg : array
        Array of average values
    """

    # if dBaverage is true, convert reflectivity to linear Z
    if dB_averaging:
        field = 10 ** (field / 10)

    # calculate background field with circular footprint
    field_bkg = scipy.ndimage.generic_filter(
        field.filled(np.nan),
        function=np.nanmean,
        mode="constant",
        footprint=bkg_mask_array.astype(bool),
        cval=np.nan,
    )

    # if calc_thres is not none, then calculate the number of points used to calculate average
    if calc_thres is not None:
        # count valid points
        field_count = scipy.ndimage.generic_filter(
            field.filled(0),
            function=np.count_nonzero,
            mode="constant",
            footprint=bkg_mask_array.astype(bool),
            cval=0,
        )
        # find threshold number of points
        val = calc_thres * np.count_nonzero(bkg_mask_array)
        # mask out values
        field_bkg = np.ma.masked_where(field_count < val, field_bkg)

    # mask where original field is invalid
    field_bkg = np.ma.masked_where(field.mask, field_bkg)

    # if dBaveraging is true, convert background reflectivity to dBZ
    if dB_averaging:
        field_bkg = 10 * np.log10(field_bkg)
        # mask where original field is invalid
        field_bkg = np.ma.masked_where(field.mask, field_bkg)

    return field_bkg


def core_cos_scheme(
    field, field_bkg, max_diff, zero_diff_cos_val, always_core_thres, CS_CORE
):
    """
    Function for assigning cores based on a cosine function

    Parameters
    ----------
    field : array
        Field values
    field_bkg : array
        Background average of field values
    max_diff : float
        Maximum difference between field and field_bkg needed for feature detection
    zero_diff_cos_val : float
        Value where the cosine function returns a zero difference
    always_core_thres : float
        All values above this threshold considered to be features
    CS_CORE : int
        Value assigned to features

    Returns
    -------
    core_array : array
        Array of booleans if point is feature (1) or not (0)
    """

    # initialize entire array to not a core
    core_array = np.zeros_like(field)

    # calculate zeDiff for entire array
    zDiff = max_diff * np.cos((np.pi * field_bkg) / (2 * zero_diff_cos_val))
    zDiff[zDiff < 0] = 0  # where difference less than zero, set to zero
    zDiff[field_bkg < 0] = max_diff  # where background less than zero, set to max. diff

    # set core values
    # where field >= always_core_thres and where difference exceeds minimum
    core_elements = np.logical_or(
        (field >= always_core_thres), (field - field_bkg) >= zDiff
    )
    core_elements = core_elements.filled(0)
    core_array[core_elements] = CS_CORE

    # mask by field array
    core_array = np.ma.masked_where(field.mask, core_array)

    return core_array


def core_scalar_scheme(
    field, field_bkg, max_diff, always_core_thres, CORE, use_addition=False
):
    """
    Function for assigning cores based on a scalar difference

    Parameters
    ----------
    field : array
        Field values
    field_bkg : array
        Background average of field values
    max_diff : float
        Maximum difference between field and field_bkg needed for feature detection
    always_core_thres : float
        All values above this threshold considered to be a feature
    CORE : int
        Value assigned to features
    use_addition : bool
        Boolean to determine if scalar should be added (True) or multiplied (False)

    Returns
    -------
    core_array : array
        Array of booleans if point is feature (1) or not (0)
    """

    # initialize entire array to not a core
    core_array = np.zeros_like(field)

    # calculate zDiff for entire array
    # if addition, add difference. Else, multiply difference
    if use_addition:
        zDiff = (max_diff + field_bkg) - field_bkg
    else:
        zDiff = (max_diff * field_bkg) - field_bkg

    zDiff[zDiff < 0] = 0  # where difference less than zero, set to zero
    zDiff[field_bkg < 0] = 0  # where background less than zero, set to zero

    # set core values
    # where field >= always_core_thres and where difference exceeds minimum
    core_elements = np.logical_or(
        (field >= always_core_thres), (field - field_bkg) >= zDiff
    )
    core_elements = core_elements.filled(0)
    core_array[core_elements] = CORE

    # mask by field array
    core_array = np.ma.masked_where(field.mask, core_array)

    return core_array


def create_radius_mask(max_diameter, radius_km, x_spacing, y_spacing, center_mask_x):
    """
    Creates a circular mask based on input diameter

    Parameters
    ----------
    max_diameter : int
        maximum diameter in kilometers
    radius_km : int
        radius in kilometers
    x_spacing, y_spacing : float
        x- and y-dimension pixel size in meters, respectively
    center_mask_x : int
        index of center point

    Returns
    -------
    feature_mask_array : array
        array masked based on distance of diameter to incorporate
    """

    feature_mask_array = np.zeros((max_diameter, max_diameter))
    feature_mask_array = create_radial_mask(
        feature_mask_array,
        0,
        radius_km,
        x_spacing,
        y_spacing,
        center_mask_x,
        center_mask_x,
        True,
    )

    return feature_mask_array


def assign_feature_radius_km(field_bkg, val_for_max_rad, max_rad=5):
    """
    Assigns the radius in kilometers based on the background values

    Parameters
    ----------
    field_bkg : array
        array of background field values
    val_for_max_rad : float
        field value for maximum radius (5 km)
    max_rad : float, optional
        maximum radius in kilometers

    Returns
    -------
    radius_array_km : array
        array of radii based on background values and val for max. radius
    """

    radius_array_km = np.ones_like(field_bkg)

    radius_array_km[field_bkg >= (val_for_max_rad - 15)] = max_rad - 3
    radius_array_km[field_bkg >= (val_for_max_rad - 10)] = max_rad - 2
    radius_array_km[field_bkg >= (val_for_max_rad - 5)] = max_rad - 1
    radius_array_km[field_bkg >= val_for_max_rad] = max_rad

    return radius_array_km


def classify_feature_array(
    field,
    feature_array,
    core_array,
    NOSFCECHO,
    FEAT_VAL,
    BKGD_VAL,
    WEAKECHO,
    CORE,
    MINDBZUSE,
    WEAKECHOTHRES,
):
    """
    Does an initial feature detection

    Parameters
    ----------
    field : array
        Array of field values
    feature_array : array
        Array with feature detection
    core_array : array
        Array with cores
    NOSFCECHO : int
        Value to assign points classified as no surface echo
    FEAT_VAL : int
        Value to assign points classified as features
    BKGD_VAL : int
        Value to assign points classified as background echo
    WEAKECHO : int
        Value to assign points classfied as weak echo
    CORE : int
        Value assigned to cores in core_array
    MINDBZUSE : float
        Minimum dBZ value to consider in classification, all values below this will be set to NOSFCECHO
    WEAKECHOTHRES : float
        dBZ threshold for weak echo classification, all values below this will be set to WEAKECHO

    Returns
    -------
    feature_array : array
        array of classifications
    """

    # assuming order so that each point is only assigned one time, no overlapping assignment
    # initially, assign every point to stratiform
    feature_array[:] = BKGD_VAL
    # where field is masked, set to no surface echo
    feature_array[field.mask] = NOSFCECHO
    # assign cores to FEAT
    feature_array[core_array == CORE] = FEAT_VAL
    # assign field values less than weakechothres to weak echo
    feature_array[field < WEAKECHOTHRES] = WEAKECHO
    # assign field values less than minimum to no surface echo
    feature_array[field < MINDBZUSE] = NOSFCECHO

    return feature_array



"""
=================================
Grid Data Using Xradar and Py-ART
=================================

An example which uses xradar and Py-ART to grid a PPI file.

"""

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause


import xradar as xd

import pyart
from pyart.testing import get_test_data

# Locate the test data and read in using xradar
filename = get_test_data("swx_20120520_0641.nc")
tree = xd.io.open_cfradial1_datatree(filename)

# Give the tree Py-ART radar methods
radar = tree.pyart.to_radar()

# Grid using 11 vertical levels, and 101 horizontal grid cells at a resolution on 1 km
grid = pyart.map.grid_from_radars(
    (radar,),
    grid_shape=(11, 101, 101),
    grid_limits=(
        (0.0, 10_000),
        (-50_000.0, 50_000.0),
        (-50_000, 50_000.0),
    ),
)

display = pyart.graph.GridMapDisplay(grid)
display.plot_grid(
    "reflectivity_horizontal", level=0, vmin=-20, vmax=60, cmap="ChaseSpectral"
)



""" Unit Tests for Py-ART's util/columnsect.py module. """

import numpy as np

import pyart

# read in example file
radar = pyart.io.read_nexrad_archive(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)


def test_get_azimuth():
    """
    test to make sure azimuth is correct to Everett, WA
    """
    azimuth = pyart.util.columnsect.for_azimuth(
        radar.latitude["data"][0], 47.97, radar.longitude["data"][0], -122.20
    )
    test_azi = abs(np.around(azimuth, 2) - 138.57)
    assert test_azi < 0.001


def test_sphere_distance():
    """
    test to make sure sphere distance is correct to Everett, WA
    """
    distance = pyart.util.columnsect.sphere_distance(
        radar.latitude["data"][0], 47.97, (radar.longitude["data"][0]), -122.20
    )
    test_dis = abs(np.around(distance / 1000.0, 2) - 33.27)
    assert test_dis < 0.001


def test_get_field_location():
    """
    test to make sure column above location is pulled correctly
    """
    column = pyart.util.columnsect.get_field_location(radar, 47.97, -122.20)

    # check to make sure z-gate is pulled correctly.
    test_height = abs(column.height[0] - 564)
    assert test_height < 0.001

    # check to make sure reflectivity value is minimum
    test_z = abs(column.reflectivity[0] + 32)
    assert test_z < 0.001


def test_column_vertical_profile():
    """
    test to make sure CVP column above location is pulled correctly
    """
    column = pyart.util.columnsect.column_vertical_profile(
        radar, 47.97, -122.20, azimuth_spread=3, spatial_spread=5
    )
    # check to make sure z-gate is pulled correctly.
    test_height = abs(column.height.data[0] - 565.7)
    assert test_height < 0.001

    # check to make sure reflectivity value is minimum
    test_z = abs(column.reflectivity.data[0] + 32)
    assert test_z < 0.001

    # check to make sure time-offset is correct
    test_offset = abs(column.time_offset.data[0] - 27.492)
    assert test_offset < 0.001
